@@ -16,6 +16,7 @@
  * Inc.,  51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  */
 
+#include <pthread.h>
 #include "platform_defs.h"
 #include "command.h"
 #include "input.h"
@@ -25,6 +26,7 @@
 char	*progname;
 int	exitcode;
 int	expert;
+int	idlethread;
 size_t	pagesize;
 struct timeval stopwatch;
 
@@ -32,7 +34,7 @@ void
 usage(void)
 {
 	fprintf(stderr,
-		_("Usage: %s [-adfmnrRstVx] [-p prog] [-c cmd]... file\n"),
+_("Usage: %s [-adfinrRstVx] [-m mode] [-p prog] [[-c|-C] cmd]... file\n"),
 		progname);
 	exit(1);
 }
@@ -56,7 +58,11 @@ init_commands(void)
 {
 	attr_init();
 	bmap_init();
+	copy_range_init();
+	cowextsize_init();
+	encrypt_init();
 	fadvise_init();
+	fiemap_init();
 	file_init();
 	flink_init();
 	freeze_init();
@@ -65,7 +71,6 @@ init_commands(void)
 	help_init();
 	imap_init();
 	inject_init();
-	seek_init();
 	madvise_init();
 	mincore_init();
 	mmap_init();
@@ -73,20 +78,27 @@ init_commands(void)
 	parent_init();
 	pread_init();
 	prealloc_init();
-	fiemap_init();
 	pwrite_init();
 	quit_init();
 	readdir_init();
+	reflink_init();
 	resblks_init();
+	seek_init();
 	sendfile_init();
 	shutdown_init();
 	sync_init();
 	sync_range_init();
 	truncate_init();
+	utimes_init();
 }
 
+/*
+ * This allows xfs_io commands specified on the command line to be run on every
+ * open file in the file table. Commands that should not be iterated across all
+ * open files need to specify CMD_FLAG_ONESHOT in their command flags.
+ */
 static int
-init_args_command(
+filetable_iterator(
 	int	index)
 {
 	if (index >= filecount)
@@ -99,9 +111,6 @@ static int
 init_check_command(
 	const cmdinfo_t	*ct)
 {
-	if (ct->flags & CMD_FLAG_GLOBAL)
-		return 1;
-
 	if (!file && !(ct->flags & CMD_NOFILE_OK)) {
 		fprintf(stderr, _("no files are open, try 'help open'\n"));
 		return 0;
@@ -138,7 +147,7 @@ init(
 	pagesize = getpagesize();
 	gettimeofday(&stopwatch, NULL);
 
-	while ((c = getopt(argc, argv, "ac:dFfmp:nrRstTVx")) != EOF) {
+	while ((c = getopt(argc, argv, "ac:C:dFfim:p:nrRstTVx")) != EOF) {
 		switch (c) {
 		case 'a':
 			flags |= IO_APPEND;
@@ -146,6 +155,9 @@ init(
 		case 'c':
 			add_user_command(optarg);
 			break;
+		case 'C':
+			add_oneshot_user_command(optarg);
+			break;
 		case 'd':
 			flags |= IO_DIRECT;
 			break;
@@ -155,6 +167,9 @@ init(
 		case 'f':
 			flags |= IO_CREAT;
 			break;
+		case 'i':
+			idlethread = 1;
+			break;
 		case 'm':
 			mode = strtoul(optarg, &sp, 0);
 			if (!sp || sp == optarg) {
@@ -206,16 +221,42 @@ init(
 	}
 
 	init_commands();
-	add_args_command(init_args_command);
+	add_command_iterator(filetable_iterator);
 	add_check_command(init_check_command);
 }
 
+/*
+ * The purpose of this idle thread is to test io from a multi threaded process.
+ * With single threaded process, the file table is not shared and file structs
+ * are not reference counted. Spawning an idle thread can help detecting file
+ * struct reference leaks.
+ */
+void *
+idle_loop(void *arg)
+{
+	for (;;)
+		pause();
+}
+
+void
+start_idle_thread()
+{
+	pthread_t t;
+
+	if (pthread_create(&t, NULL, idle_loop, NULL)) {
+		fprintf(stderr, "Error creating idle thread\n");
+		exit(1);
+	}
+}
+
 int
 main(
 	int	argc,
 	char	**argv)
 {
 	init(argc, argv);
+	if (idlethread)
+		start_idle_thread();
 	command_loop();
 	return exitcode;
 }
